@@ -407,6 +407,11 @@ use PlSense::Entity::Array;
                 VALUE:
                 foreach my $value ( @{$idxh->{$idx}} ) {
                     $self->add_substitute($curraddr, $value);
+
+                    # Add reverse route if value is a argument to super method
+                    if ( eval { $value->isa("PlSense::Entity") } ) { next VALUE; }
+                    if ( $mtdnm !~ m{ \A SUPER:: }xms ) { next VALUE; }
+                    addrrouter->add_reverse_route($value, $curraddr);
                 }
             }
             delete $unknownargh_of{ident $self}->{$mtdaddr};
